;;;;
;;;; lisp2triples.lisp
;;;;
;;;; George Ferguson, ferguson@cs.rochester.edu, 28 Aug 2002
;;;; Time-stamp: <Thu Sep  5 15:28:09 EDT 2002 ferguson>
;;;;
;;;; Converts FOL formula expressed in KIF (or KIF-like Lisp anyway)
;;;; to RDF triples for use in DAML.
;;;;
;;;; Based on ``A Proposal for Encoding Logic in RDF/DAML'',
;;;; Draft of August 8, 2001, by Drew McDermott, Jonathan Borden, Mark
;;;; Burstein, Doug Smith, and Richard Waldinger.
;;;;
;;;; NOTES:
;;;; - To really translate KIF, need to handle non-binary AND/OR, ==>, etc.
;;;; - We assume that any nested forms are functional terms, not atomic
;;;;   formulas nested inside other predicates. That is, we assume we are
;;;;   translating FOL.
;;;; - The EXISTS! quantifier isn't part of KIF, but we let it through.
;;;;

;;;
;;; Packages needed to correspond to ``drs'' and ``rdf'' namespaces
;;;
;;; Note that the ``rdf_i'' symbols will be interned on the fly, but
;;; they need to be listed and exported here for use by code that runs
;;; in an image that didn't run Trtop.
;;;
(defpackage drs
  (:export Or And Implies Forall Exists Exists! Not Var Triple
	   args antecedent consequent vars type body seqtype funapp
	   toplevel)
  (:shadow type))

(defpackage rdf
  (:export type Bag subj pred obj Seq
	   _1 _2 _3 _4 _5 _6 _7 _8 _9 _10)
  (:shadow type))

;;;
;;; Main entry point: Tr function
;;;
(defun Tr (e v q)
  "Tr(e, v, q) is the translation of expression e, which occurs inside an
environment binding variables in the list v, and whose ``syntactic path'' is q.
The value of Tr(e, v, q) is a pair <d, S>, where S is a set of triples and d
is the identifier used in that set to refer to the translation of e."
  (case (car e)
   (or
    (Tr-connection 'drs:Or (second e) (third e) v q))
   (and
    (Tr-connection 'drs:And (second e) (third e) v q))
   ((implies =>)
    (Tr-implication (second e) (third e) v q))
   ((implied-by <=)
    (Tr-implication (third e) (second e) v q))
   ((iff <=>)
    (let ((p1 (second e))
	  (p2 (third e)))
      (Tr (list 'and (list 'implies p1 p2) (list 'implies p2 p1)) v q)))
   (forall
    (Tr-quantification 'drs:Forall (second e) (third e) v q))
   (exists
    (Tr-quantification 'drs:Exists (second e) (third e) v q))
   (exists!
    (Tr-quantification 'drs:Exists! (second e) (third e) v q))
   (not
    (Tr-not (second e) v q))
   (otherwise
    (Tr-formula (car e) (cdr e) v q))))

(defun Trtop (E &optional q)
  "Toplevel translations add the toplevel attribute."
  ;; Note: Could use alternate technique detailed in paper."
  (multiple-value-bind (d tt) (Tr E nil q)
    (values d (union tt (list (triple d 'drs:toplevel 'true))))))
      

(defun Tr-connection (connective P1 P2 v q)
  ;; NOTE: P1 and P2 are P and Q in the paper, but that conflicts with ``q''
  (let ((d (gensym-id q))
	(b (gensym-id (extend q 0))))
    (multiple-value-bind (d1 t1) (Tr P1 v (extend q 1))
     (multiple-value-bind (d2 t2) (Tr P2 v (extend q 2))
      (values d (union+ (list (triple d 'rdf:type connective)
			      (triple d 'drs:args b)
			      (triple b 'rdf:_1 d1)
			      (triple b 'rdf:_2 d2))
			t1 t2))))))

(defun Tr-implication (P1 P2 v q)
  ;; NOTE: Ditto on P1/P2 vs. P/Q
  (let ((d (gensym-id q)))
    (multiple-value-bind (d1 t1) (Tr P1 v (extend q 1))
     (multiple-value-bind (d2 t2) (Tr P2 v (extend q 2))
      (values d (union+ (list (triple d 'rdf:type 'drs:Implies)
			      (triple d 'drs:antecedent d1)
			      (triple d 'drs:consequent d2))
			t1 t2))))))

(defun Tr-quantification (quantifier vars P v q)
  "VARS is a list of variables being bound by the quantifier, either just
the variable or a two-element list consisting of variable and type."
  (let* ((n (length vars))
	 (x (mapcar #'safe-car vars))
	 (S (mapcar #'safe-cadr vars))
	 (d (gensym-id q))
	 (b (gensym-id (extend q 0)))
	 (l (loop for i from 1 to n
		  collect (gensym-id (extend (extend q 1) i)))))
    ;; Note: ``tt'' is just ``t'' in the paper, but we can't have that
    (multiple-value-bind (p tt)
       (Tr P (append (mapcar #'cons x l) v) (extend q 2))
      (values d (union+ (list (triple d 'rdf:type quantifier)
			      (triple d 'drs:vars b)
			      ;; Paper example uses Seq once and Bag once...
			      (triple b 'rdf:type 'rdf:Bag))
			;; NOTE: Added name to variable attributes to support
			;; reconversion into markup as in paper example
			(reduce #'union (loop with li and Si and xi and _i
					    for i from 1 to n
					    do (setq li (nth (1- i) l))
					       (setq Si (nth (1- i) S))
					       (setq xi (nth (1- i) x))
					       (setq _i (rdftag i))
					    collect (list (triple b _i li)
							  (triple li 'rdf:type 'drs:Var)
							  (triple li 'name xi)
							  (triple li 'drs:type Si))))
			(list (triple d 'drs:body p))
			tt)))))

(defun Tr-not (P v q)
  ;; Not given in paper, but works like AND/OR with just one arg
  (let ((d (gensym-id q))
	(b (gensym-id (extend q 0))))
    (multiple-value-bind (d1 t1) (Tr P v (extend q 1))
      (values d (union+ (list (triple d 'rdf:type 'drs:Not)
			      (triple d 'drs:args b)
			      (triple b 'rdf:_1 d1))
			t1)))))

(defun Tr-formula (P args v q)
  (when (eql (length args) 0)
    (error "Can't translate formula involving 0-ary predicate: ~A" P))
  (let* ((A (first args))
	 (B (if (eql (length args) 2) (second args) (cons 'drs:args (rest args))))
	 (d (gensym-id q)))
    (multiple-value-bind (p tp) (Tr-term P v (extend q 0))
      (multiple-value-bind (a ta) (Tr-term A v (extend q 1))
	(multiple-value-bind (b tb) (Tr-term B v (extend q 2))
	  ;; The paper says "an atomic formula is represented by four triples,
	  ;; specifying that it exists, and what its subject, predicate, and
	  ;; object are," but the code only creates three triples, and the
	  ;; example uses a <drs:Triple> tag. I guess we should add that.
	  (values d (union+ (list (triple d 'rdf:type 'drs:Triple)
				  (triple d 'rdf:subj a)
				  (triple d 'rdf:pred p)
				  (triple d 'rdf:obj b))
			    ta tp tb)))))))

(defun Tr-term (x v q)
  ;; TODO: Need URI case from paper?
  (cond
   ((listp x)
    (Tr-function x v q))
   ((assoc x v)
    (Tr-variable x v q))
   (t
    x)))

(defun Tr-function (func v q)
  (let* ((F (first func))
	 (A (rest func))
	 (k (length A))
	 (d (gensym-id q)))
    (multiple-value-bind (f t0) (Tr-term F v (extend q 0))
     (let ((at (loop with Ai
		   for i from 1 to k
		   do (setq Ai (nth (1- i) A))
		   collect (multiple-value-list (Tr-term Ai v (extend q i))))))
       (values d (union+ (list (triple d 'rdf:type 'rdf:Seq)
			       (triple d 'drs:seqtype 'drs:funapp)
			       (triple d 'rdf:_1 f))
			 (loop with ai
			     for i from 1 to k
			     do (setq ai (first (nth (1- i) at)))
			     ;; Error in paper: has _i here
			     collect (triple d (rdftag (1+ i)) ai))
			 (reduce #'union (cons t0 (loop with ti
						      for i from 1 to k
						      do (setq ti (second (nth (1- i) at)))
						      collect ti)))))))))

(defun Tr-variable (u v q)
  "Returns the first entry in V whose first element is U (actually, returns
the CDR of that element, ie., the translation of the variable)."
  (declare (ignore q))
  (cdr (assoc u v)))

;;;
;;; Utility functions
;;;

(defun extend (path i)
  "Returns a new path with is PATH extended with the element I. This function
is called ``++'' in the paper, but this is clearer for the code."
  (append path (list i)))

(defun gensym-id (path)
  "Creates a new symbol ``G_i1_i2_..._ik'' where the i's are the elements
of PATH. This function is called ``gensym'' in the paper, but that is
(obviously) a reserved word in Common Lisp."
  (let ((s "G"))
    (loop for i in path
	  do (setq s (format nil "~A_~D" s i)))
    (intern s)))

(defun triple (x y z)
  "Returns a new triple whose components are X, Y, and Z."
  (list x y z))

(defun rdftag (i)
  "Returns the symbol ``rdf:_i'' for given ``i''."
  (let* ((package (find-package :rdf))
	 (sym (intern (format nil "_~D" i) (find-package :rdf))))
    (export sym package)
    sym))

(defun safe-car (x)
  "Returns (CAR X) if X is a cons, otherwise returns X."
  (if (consp x)
      (car x)
    x))

(defun safe-cadr (x)
  "Returns (CADR X) if X and (CDR X) are conses, otherwise returns NIL."
  (when (and (consp x) (consp (cdr x)))
    (cadr x)))

(defun union+ (&rest sets)
  "Returns the union of the given SETS."
  (reduce #'union sets))

;;;
;;; Debugging
;;;

#||
(trace
 Tr-connection
 Tr-implication
 Tr-quantification
 Tr-formula
 Tr-term
 Tr-function
 Tr-variable
 ;; extend
 ;; gensym-id
 ;; rdftag
 ;; triple
 ;; safe-car
 ;; safe-cadr
)
||#

;;;
;;; Testing
;;;

(defun test1 ()
  "Example from paper:
    (A x)[[(E y) R(y,x)] => P(x,f(x),g(x))]

Note that example markup has extra instance of x (aka, g256) in the args
used as the obj of predicate P (the x is actually the subj, and the args
are the other two arguments, if you follow me."
  (let ((expr '(forall (x)
		       (implies (exists (y) (R y x))
				(P x (f x) (g x))))))
    (Trtop expr)))

