;;;;
;;;; triples2daml.lisp
;;;;
;;;; George Ferguson, ferguson@cs.rochester.edu,  4 Sep 2002
;;;; Time-stamp: <Thu Sep  5 15:56:51 EDT 2002 ferguson>
;;;;
;;;; Requires at least the package definitions from lisp2triples.lisp.
;;;;
;;;; Function XMLIFY converts set of triples to XML entities
;;;; represented using CLOS objects. These objects print themselves
;;;; using XML syntax.
;;;;

(defvar *xml-print-indent* 0
  "Used to store current indent during XML-ENTITY print functions.")

(defvar *xml-print-indent-incr* 1
  "Amount to indent nested entities relative to their parent.")

(defun tabulate (n stream)
  "Print N spaces to STREAM."
  (loop for i from 0 below n do (princ #\Space stream)))

(defun iformat (stream fmt &rest args)
  "Tab to current indentation level and call FORMAT."
  (tabulate *xml-print-indent* stream)
  (apply #'format stream fmt args))

(defun iformat+ (stream fmt &rest args)
  "Tab to current indentation level, call FORMAT, then increment the
current indentation level by one increment."
  (tabulate *xml-print-indent* stream)
  (apply #'format stream fmt args)
  (incf *xml-print-indent* *xml-print-indent-incr*))

(defun iformat- (stream fmt &rest args)
  "Decrement current indentation level, tab to current indentation level,
then call FORMAT."
  (decf *xml-print-indent* *xml-print-indent-incr*)
  (tabulate *xml-print-indent* stream)
  (apply #'format stream fmt args))

;;;
;;; Classes used to hold XML/RDF/DAML entities (and their print functions)
;;;
(defclass xml-entity ()
  ((type :initarg :type)
   (id :initarg :id)
   (toplevel :initarg :toplevel :initform nil))
  (:documentation "Base class for all translated entities."))

(defclass connection (xml-entity)
  ((arg1 :initarg :arg1)
   (arg2 :initarg :arg2))
  (:documentation "Two-argument connectives like AND, OR, etc."))

(defmethod print-object ((this connection) stream)
  (with-slots (type id toplevel arg1 arg2) this
    (when toplevel
      (setq *xml-print-indent* 0))
    (let ((type (string-capitalize (string type))))
      (iformat+ stream "<drs:~A~:[~; id=\"~A\" toplevel=\"true\"~]>~%" type toplevel id)
      (iformat+ stream "<li>~%")
      (print-object arg1 stream)
      (iformat- stream "</li>~%")
      (iformat+ stream "<li>~%")
      (print-object arg2 stream)
      (iformat- stream "</li>~%")
      (iformat- stream "</drs:~A>~%" type))))

(defclass implication (xml-entity)
  ((antecedent :initarg :antecedent)
   (consequent :initarg :consequent))
  (:documentation "Implications like IMPLIES, IMPLIED-BY, IFF, etc."))

(defmethod print-object ((this implication) stream)
  (with-slots (id toplevel antecedent consequent) this
    (when toplevel
      (setq *xml-print-indent* 0))
    (iformat+ stream "<drs:Implies~:[~; id=\"~A\" toplevel=\"true\"~]>~%" toplevel id)
    (iformat+ stream "<drs:antecedent>~%")
    (print-object antecedent stream)
    (iformat- stream "</drs:antecedent>~%")
    (iformat+ stream "<drs:consequent>~%")
    (print-object consequent stream)
    (iformat- stream "</drs:consequent>~%")
    (iformat- stream "</drs:Implies>~%")))

(defclass quantification (xml-entity)
  ((vars :initarg :vars)
   (body :initarg :body))
  (:documentation "Quantifying forms like FORALL, EXISTS, etc."))

(defmethod print-object ((this quantification) stream)
  (with-slots (type id toplevel vars body) this
    (when toplevel
      (setq *xml-print-indent* 0))
    (let ((type (string-capitalize (string type))))
      (iformat+ stream "<drs:~A~:[~; id=\"~A\" toplevel=\"true\"~]>~%" type toplevel id)
      (iformat+ stream "<drs:vars>~%")
      (print-object vars stream)
      (iformat- stream "</drs:vars>~%")
      (iformat+ stream "<drs:body>~%")
      (print-object body stream)
      (iformat- stream "</drs:body>~%")
      (iformat- stream "</drs:~A>~%" type))))

(defclass negation (xml-entity)
  ((body :initarg :body))
  (:documentation "Negations."))

(defmethod print-object ((this negation) stream)
  (with-slots (toplevel	id body) this
    (when toplevel
      (setq *xml-print-indent* 0))
    (iformat+ stream "<drs:Not~:[~; id=\"~A\" toplevel=\"true\"~]>~%" toplevel id)
    (print-object body stream)
    (iformat- stream "</drs:Not>~%")))

(defclass bag (xml-entity)
  ((elements :initarg :elements))
  (:documentation "Unordered sets."))

;; Note: Bag's are only used for bound variables in a quantifier,
;; so we specialize the printing to match the example in the paper
;; (in particular, <li><drs:Var ...></li> for each variable).
(defmethod print-object ((this bag) stream)
  (with-slots (elements) this
    (iformat+ stream "<drs:Bag>~%")
    (loop for e in elements
	do (iformat stream "<li>")
	   (print-object e stream)
	   (format stream "</li>~%"))
    (iformat- stream "</drs:Bag>~%")))

(defclass seq (xml-entity)
  ((seqtype :initarg :seqtype)
   (elements :initarg :elements))
  (:documentation "Ordered lists."))

(defmethod print-object ((this seq) stream)
  (with-slots (seqtype elements) this
    (iformat+ stream "<drs:Seq seqtype=\"~A\">~%" seqtype)
    (loop for e in elements
	do (cond
	    ((typep e 'seq)
	     (iformat+ stream "<li>~%")
	     (print-object e stream)
	     (iformat- stream "</li>~%"))
	     (t
	      (iformat stream "<li resource=")
	      (print-object e stream)
	      (format stream "/>~%"))))
    (iformat- stream "</drs:Seq>~%")))

(defclass triple (xml-entity)
  ((subj :initarg :subj)
   (pred :initarg :pred)
   (obj :initarg :obj))
  (:documentation "Triples (used for atomic formulas)."))

(defmethod print-object ((this triple) stream)
  (with-slots (id toplevel subj pred obj) this
    (when toplevel
      (setq *xml-print-indent* 0))
    (iformat+ stream "<drs:Triple~:[~; id=\"~A\" toplevel=\"true\"~]>~%" toplevel id)
    (cond
     ((typep subj 'seq)
      (iformat+ stream "<drs:obj>~%")
      (print-object subj stream)
      (iformat- stream "</drs:obj>~%"))
     (t
      (iformat stream "<obj resource=")
      (print-object subj stream)
      (format stream "/>~%")))
    (iformat stream "<pred resource=")
    (print-object pred stream)
    (format stream "/>~%")
    (cond
     ((typep obj 'seq)
      (iformat+ stream "<drs:obj>~%")
      (print-object obj stream)
      (iformat- stream "</drs:obj>~%"))
     (t
      (iformat stream "<obj resource=")
      (print-object obj stream)
      (format stream "/>~%")))
    (iformat- stream "</drs:Triple>~%")))

(defclass var (xml-entity)
  ((name :initarg :name))
  (:documentation "Variables"))

;; Note: Since always printed inside Bag, don't insert indentation (iformat)
(defmethod print-object ((this var) stream)
  (with-slots (type id name) this
    (if type 
	(format stream "<drs:Var id=\"~A\" type=\"~A\" name=\"~A\"/>" id type name)
      (format stream "<drs:Var id=\"~A\" name=\"~A\"/>" id name))))
    
(defclass ref (xml-entity)
  ((resource :initarg :resource))
  (:documentation "References to bound vars."))

(defmethod print-object ((this ref) stream)
  (with-slots (resource) this
    (format stream "\"#~A\"" resource)))
    
;;;
;;; Main entry point:
;;;
(defun xmlify (triples)
  "Returns a list of entities, one for each toplevel entity described
by the set of TRIPLES."
  (mapcar #'(lambda (x)
	      (let ((obj (xmlify-entity x triples)))
		(setf (slot-value obj 'toplevel) t)
		obj))
	  (get-subjs-by-pred triples 'drs:toplevel)))

(defun xmlify-entity (entity triples)
  (let ((type (get-type-of-entity entity triples)))
    (case type
      ((drs:And drs:Or)
       (xmlify-connection type entity triples))
      (drs:Implies
       (xmlify-implication entity triples))
      ((drs:Forall drs:Exists drs:Exists!)
       (xmlify-quantification type entity triples))
      (drs:Not
       (xmlify-negation entity triples))
      (rdf:Seq
       (xmlify-seq entity triples))
      (drs:Triple
       (xmlify-triple entity triples))
      (t
       (xmlify-other entity type triples)))))

(defun get-type-of-entity (entity triples)
  (get-obj-by-subj-and-pred triples entity 'rdf:type))

(defun xmlify-connection (connective entity triples)
  (let ((args (get-obj-by-subj-and-pred triples entity 'drs:args)))
    (make-instance 'connection
      :type connective
      :id entity
      :arg1 (xmlify-entity (get-obj-by-subj-and-pred triples args 'rdf:_1) triples)
      :arg2 (xmlify-entity (get-obj-by-subj-and-pred triples args 'rdf:_2) triples))))

(defun xmlify-implication (entity triples)
  (make-instance 'implication
    :id entity
    :antecedent (xmlify-entity (get-obj-by-subj-and-pred triples entity 'drs:antecedent) triples)
    :consequent (xmlify-entity (get-obj-by-subj-and-pred triples entity 'drs:consequent) triples)))

(defun xmlify-quantification (quantifier entity triples)
  (make-instance 'quantification
    :type quantifier
    :id entity
    :vars (xmlify-vars (get-obj-by-subj-and-pred triples entity 'drs:vars) triples)
    :body (xmlify-entity (get-obj-by-subj-and-pred triples entity 'drs:body) triples)))

(defun xmlify-negation (entity triples)
  (let ((args (get-obj-by-subj-and-pred triples entity 'drs:args)))
    (make-instance 'negation
      :id entity
      :body (xmlify-entity (get-obj-by-subj-and-pred triples args 'rdf:_1) triples))))

(defun xmlify-vars (entity triples)
  (make-instance 'bag
    :id entity
    :elements
    (loop with _i and obj
	for i from 1
	do (setq _i (rdftag i))
	   (setq obj (get-obj-by-subj-and-pred triples entity _i))
	while obj
	collect (xmlify-var-def obj triples))))

(defun xmlify-var-def (entity triples)
  (make-instance 'var
    :id entity
    :type (get-obj-by-subj-and-pred triples entity 'drs:type)
    :name (get-obj-by-subj-and-pred triples entity 'name)))

(defun xmlify-seq (entity triples)
  (make-instance 'seq
    :id entity
    :seqtype (get-obj-by-subj-and-pred triples entity 'drs:seqtype)
    :elements
    (loop with _i and obj
	for i from 1
	do (setq _i (rdftag i))
	   (setq obj (get-obj-by-subj-and-pred triples entity _i))
	while obj
	collect (xmlify-entity obj triples))))

(defun xmlify-triple (entity triples)
  (make-instance 'triple
    :id entity
    :subj (xmlify-entity (get-obj-by-subj-and-pred triples entity 'rdf:subj) triples)
    :pred (xmlify-entity (get-obj-by-subj-and-pred triples entity 'rdf:pred) triples)
    :obj (xmlify-entity (get-obj-by-subj-and-pred triples entity 'rdf:obj) triples)))

(defun xmlify-other (entity type triples)
  (case type
    (drs:Var
     (make-instance 'ref :resource entity))
    (t
     (format nil "~S" entity))))

;;;
;;; Functions for accessing triples
;;;
(defun get-triples-by-subj (triples subj)
  "Returns a list of those TRIPLES whose subject (first parameter) is SUBJ."
  (mapcan #'(lambda (triple)
	      (when (eq (first triple) subj)
		(list triple)))
	  triples))

(defun get-triples-by-pred (triples pred)
  "Returns a list of those TRIPLES whose predicate (second parameter) is PRED."
  (mapcan #'(lambda (triple)
	      (when (eq (second triple) pred)
		(list triple)))
	  triples))

(defun get-subjs-by-pred (triples pred)
  "Returns a list of the subjects (first parameters) of any TRIPLES whose
predicate (second parameter) is PRED."
  (mapcar #'first (get-triples-by-pred triples pred)))

(defun get-triples-by-subj-and-pred (triples subj pred)
  "Returns a list of those TRIPLES whose subject (first parameter) is SUBJ
and whose predicate (second parameter) is PRED."
  (mapcan #'(lambda (triple)
	      (when (and (eq (first triple) subj) (eq (second triple) pred))
		(list triple)))
	  triples))

(defun get-obj-by-subj-and-pred (triples subj pred)
  "Returns the obj (third parameter) of the first triple in TRIPLES whose
subject (first parameter) is SUBJ and whose predicate (second parameter)
is PRED."
  (third (first (get-triples-by-subj-and-pred triples subj pred))))

;;;
;;; Debuygging
;;;
#||
(trace
 xmlify
 xmlify-entity
 get-type-of-entity
 xmlify-connection
 xmlify-implication
 xmlify-quantification
 xmlify-negation
 xmlify-seq
 xmlify-vars
 xmlify-var-def
 xmlify-other
 ;; get-triples-by-subj
 ;; get-triples-by-pred
 get-subjs-by-pred
 get-triples-by-subj-and-pred
 get-obj-by-subj-and-pred
 )
||#

;;;
;;; Testing
;;;

(defun test2 ()
  "Continue example from paper in lisp2triples.lisp. See note there
regarding error in example."
  (multiple-value-bind (name triples) (test1)
    (xmlify triples)))
